/**
 * @package css-modules-types-loader
 * @license MIT
 * @version 0.3.7
 * @author undefined <undefined>
 * @description A webpack loader that dynamically generates TypeScript typings for CSS modules from css-loader.
 * @see https://github.com/nuintun/css-modules-types-loader#readme
 */

import fs from 'fs';
import { EOL } from 'os';
import { parse } from 'acorn';
import { promisify } from 'util';
import { simple } from 'acorn-walk';

/**
 * @module utils
 */
const rm = promisify(fs.rm);
const writeFile = promisify(fs.writeFile);
async function removeFile(path) {
  if (fs.existsSync(path)) {
    await rm(path);
  }
}
function isString(value) {
  return Object.prototype.toString.call(value) === '[object String]';
}
function collect(styles, left, right) {
  if (right.type === 'Literal') {
    const value = right.value;
    if (isString(value)) {
      let key;
      switch (left.type) {
        case 'Identifier':
          key = left.name;
          break;
        case 'Literal':
          key = left.value;
          break;
      }
      if (isString(key)) {
        styles.push([key, value]);
      }
    }
  }
}
function parseStyles(content) {
  let named = false;
  const styles = [];
  const ast = parse(content, {
    sourceType: 'module',
    ecmaVersion: 'latest'
  });
  simple(ast, {
    ExpressionStatement(node) {
      simple(node, {
        AssignmentExpression({ left }) {
          if (left.type === 'MemberExpression') {
            const { property } = left;
            if (property.type === 'Identifier') {
              if (property.name === 'locals') {
                simple(node, {
                  Property({ key, value }) {
                    collect(styles, key, value);
                  }
                });
              }
            }
          }
        }
      });
    },
    ExportNamedDeclaration(node) {
      named = true;
      simple(node, {
        VariableDeclarator({ id, init }) {
          if (init) {
            collect(styles, id, init);
          }
        }
      });
    }
  });
  return [styles, named];
}
function generateTypings(content, banner, eol = EOL) {
  const [styles, named] = parseStyles(content);
  if (styles.length > 0) {
    const typings = banner ? [banner, ''] : [];
    if (named) {
      for (const [key, value] of styles) {
        typings.push(`export const ${key}: ${JSON.stringify(value)};`);
      }
    } else {
      typings.push('declare const styles: {');
      for (const [key, value] of styles) {
        typings.push(`  ${key}: ${JSON.stringify(value)}`);
      }
      typings.push('};', '', 'export default styles;');
    }
    return typings.join(eol);
  }
  return null;
}

export { collect, generateTypings, isString, parseStyles, removeFile, rm, writeFile };
